AttachSpec("lat.spec");
Attach("AutHyp.m");
//SetDebugOnError(true);
Q:= Rationals();
Z:= Integers();
R:= QNF();

ModZ := func< x | x - Floor(x) >;
Mod2Z:= func< x | x - Floor(x/2)*2 >;

Norms:= function(L, p)
  D:= Dual(L);
  DZ, h1:= ZLattice(D);
  Q, h:= quo< DZ | ZLattice(L) >;
  f:= h1*h;
  if p ne 0 then Q:= SylowSubgroup(Q, p); end if;
  return {* Rationals() | Mod2Z(InnerProduct(x,x)) where x:= D ! (q @@ f) : q in Q *};
end function;

function DiscAut(F)
  F:= Matrix(Q, F);
  FI:= F^-1;
  L:= StandardLattice( Ncols(F) );
  M:= LatticeWithBasis(F^-1);
  I, f:= quo< M | L >;

  J:= [ [] ];
  for i in [1..Ngens(I)] do
    n:= Mod2Z( (x*F, x) ) where x:= I.i @@ f;
    Y:= [ y : i in I | Mod2Z( (y*F,y) ) eq n where y:= i @@ f ];
    S:= [ ModZ( ( (I.j@@f)*F, (I.i@@f) ) ) : j in [1..i-1] ];
    if i eq 1 then 
      J:= [ [y]: y in Y ];
    else
      JJ:= [];
      for j in J do
        jF:= Matrix(j) * F;
        JJ cat:= [ Append(j, y) : y in Y | [ ModZ(e) : e in Eltseq(tmp) ] eq S where tmp:=  jF * Matrix(1, Eltseq(y)) ];
      end for;
      J:= JJ;
    end if;
  end for;

  return I, f, [ < t @ f : t in x > : x in J ];
end function;

function MyIso( A, Gens, Images )
  AGens:= ChangeUniverse(Gens, A);
  G:= Universe(A);
  Y:= [];
  for i in [1..Ngens(A)] do
    x, b:= Representation( AGens, A ! G.i );
    assert b eq 1;
    Append(~Y, &+[ G | x[j] * Images[j]: j in [1..#x] ]);
  end for;
  f:= iso< G -> G | Y >;
  assert Images eq [ f(g) : g in Gens ];
  return f;
end function;

function MyStabGens(G, h, X)
  if ISA(Type(G), Grp) then
    n:= Ngens(G);
    Gens:= [G.i: i in [1..n] ];
  else
    Gens:= G;
    n:= #G;
    G:= Universe(G);
  end if;
  Path:= [ Id(G) ];
  Orbit:= {@ X @};
  Stab:= [];

  i:= 0;
  while i lt #Orbit do
    i+:= 1;
    for j in [1..n], e in [-1,1] do
      g:= Gens[j]^e;
      Xx:= h(Orbit[i], g);
      idx:= Index(Orbit, Xx);
      x:= Path[i] * g;
      if idx eq 0 then
        Include(~Orbit, Xx);
	Append(~Path, x);
      else
        x:= x * Path[idx]^-1;
	assert h(X, x) eq X;
	if x ne Path[1] and x notin Stab and x^-1 notin Stab then
	  Append(~Stab, x);
	end if;
      end if;
    end for;
  end while;

  return Stab, Orbit, Path;
end function;

function LatexName(Name)
  S:= "";
  B:= Split(Name, " ");
  i:= 1;
  while i le #B do
    b:= B[i];
    c:= #[ j: j in [i..#B] | B[j] eq b ];
    if i ne 1 then S *:= "+"; end if;
    if c ne 1 then
      S *:= IntegerToString(c);
    end if;
    S *:= b[1] * "_{" * b[2..#b] * "}";
    i +:= c;
  end while;
  return S;
end function;

  function ConstructLattice(Name : Ext:= [])
    B:= Split(Name, " ");
    Lats:= [];
    for b in B do
      n:= StringToInteger(b[2..#b]);
      L:= LatticeWithGram( GramMatrix(Lattice(b[1], n)));
      Append(~Lats, L);
    end for;
    L:= OrthogonalSum(Lats);
    Q,D,H := DualQuotient(L);

    // Maps my generators to Q
    Gens:= [ ];
    i:= 0; n:= Degree(L);
    for l in Lats do
      q,d,h := DualQuotient(l);
      m:= Degree(l);
      Append(~Gens, [ Vector([0^^i] cat Eltseq(x @@ h) cat [0^^(n-m-i)]) @ H : x in Generators(q) ] );
      i +:= m;
    end for;

    // Map my automorphism into the group A:
    // This is great fun: Magma's way of expressing elements in abelian groups as words in some user defined generators is via "generic abelian groups"!
    A:= AutomorphismGroup(Q);
    QA:= GenericAbelianGroup(Q);		// why?
    I:= &cat Gens;
    S:= [ A | ];
    for i in [1..#Gens] do
      assert #Gens[i] le 2;
      if #Gens[i] eq 2 then // the D4 case:
	idx1:= Index(I, Gens[i,1]); assert idx1 ne 0;
	idx2:= Index(I, Gens[i,2]); assert idx2 ne 0;
	// transposition
	Images:= I;
	Images[idx1]:= Gens[i,2];
	Images[idx2]:= Gens[i,1];
	Append(~S, MyIso(QA, I, Images) );
	// 3-cycle
	Images:= I;
	Images[idx1]:= Gens[i,2];
	Images[idx2]:= &+Gens[i];
	Append(~S, MyIso(QA, I, Images) );
      else
	Images:= I;
	idx:= Index(I, Gens[i,1]); assert idx ne 0;
	Images[idx]*:= -1;
	Append(~S, MyIso(QA, I, Images) );
      end if;
      // add a two-cycle, if Lats[i] eq Lats[i-1]
      ok:= exists(j){j: j in [1..i-1] | Lats[i] cmpeq Lats[j] };
      if ok then
	// this only happens for An!
	assert #Gens[i] eq #Gens[j];
	Images:= I;
	for k in [1..#Gens[i]] do
	  idx1:= Index(I, Gens[i,k]); assert idx1 ne 0;
	  idx2:= Index(I, Gens[j,k]); assert idx2 ne 0;
	  Images[idx1]:= Gens[j,k];
	  Images[idx2]:= Gens[i,k];
	end for;
	Append(~S, MyIso(QA, I, Images) );
      end if;
    end for;

    if Ext eq [] then return L, H, sub< A | S >; end if;

    LL:= ext< L | Ext >;
    QQ,DD,HH:= DualQuotient(LL);
    E:= sub< Q | [ (D ! x) @ H : x in Ext ] >;
    T, O:= MyStabGens( sub< A | S >, func< o, g | g(o) >, E );
    AA:= AutomorphismGroup(QQ);
    T:= [ iso< QQ -> QQ | [ QQ.i @@ HH @ H @ f @@ H @ HH : i in [1..Ngens(QQ)] ] > : f in T ];

    return LL, HH, sub< AA | T >;
  end function;



U:= Matrix(Q, 2, [ 0,1,1,0]);
A:= [* Matrix(Q, GramMatrix(Lattice("A", i))) : i in [1..13] *];
D4:= Matrix(Q, GramMatrix(Lattice("D", 4)));
E6:= Matrix(Q, GramMatrix(Lattice("E", 6)));
E7:= Matrix(Q, GramMatrix(Lattice("E", 7)));
E8:= Matrix(Q, GramMatrix(Lattice("E", 8)));
FF:= DiagonalJoin(<U,U,U, -E8, -E8>);	 // Rang=22, det=-1, Hasse(2)=-1, negative EW:19

Names:= [
  " D4 A13 ",
  " D4 A12 A1 ",
  " D4 A11 A2 ",
  " D4 A11 A1 A1 ",
  " D4 A10 A2 A1 ",
  " D4 A9 A4 ",
  " D4 A9 A3 A1 ",
  " D4 A9 A2 A2 ",
  " D4 A9 A2 A1 A1 ",
  " D4 A8 A5 ",
  " D4 A8 A4 A1 ",
  " D4 A7 A4 A1 A1 ",
  " D4 A7 A3 A2 A1 ",
  " D4 A7 A2 A2 A1 A1 ",
  " D4 A6 A6 A1 ",
  " D4 A6 A5 A2 ",
  " D4 A6 A4 A2 A1 ",
  " D4 A6 A3 A2 A2 ",
  " D4 A5 A5 A3 ",
  " D4 A5 A5 A1 A1 A1 ",
  " D4 A5 A4 A4 ",
  " D4 A5 A4 A3 A1 ",
  " D4 A5 A3 A3 A1 A1 ",
  " D4 A4 A4 A2 A2 A1 ",
  " D4 A3 A3 A3 A2 A2 "
];

Ext:= [ []^^#Names ];
// Jetzt fügen wir noch ein paar anisotrope Vektoren hinzu:
Ext[ 3]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2,0,1/2,0,1/2,0,1/2,  0,0]) ];
Ext[ 4]:= Ext[3];
Ext[ 7]:= [Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2,0,1/2,0,1/2, 0,0,0, 1/2]) ];
Ext[ 9]:= Ext[7];
Ext[12]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2,0,1/2, 0,0,0,0, 1/2,1/2]) ];
Ext[13]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2,0,1/2, 1/2,0,1/2, 0,0, 0]) ];
Ext[14]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2,0,1/2, 0,0, 0,0, 1/2, 1/2]) ];
Ext[19]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2, 1/2,0,1/2,0,1/2, 0,0,0]) ];
Ext[20]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2, 0,0,0,0,0, 1/2,1/2,1/2]), Vector([1/2,0,0,1/2, 0,0,0,0,0, 1/2,0,1/2,0,1/2, 1/2,1/2,1/2]) ];
Ext[22]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2,0,1/2, 0,0,0,0, 1/2,0,1/2, 1/2]) ];
Ext[23]:= [ Vector([1/2,1/2,0,0, 0,0,0,0,0, 1/2,0,1/2, 1/2,0,1/2, 1/2, 1/2]), Vector([0,1/2,0,1/2, 1/2,0,1/2,0,1/2, 0,0,0, 1/2,1,1/2, 0, 1/2]) ];
Ext[25]:= [ Vector([1/2,1/2,0,0, 1/2,0,1/2, 1/2,0,1/2, 1/2,0,1/2, 0,0, 0,0]) ];


ValDiv:= func< L, p | [ Valuation(e,p) : e in ElementaryDivisors(Matrix(Z, GramMatrix(L))) | e ne 0 and e mod p eq 0 ] >;
Filter:= func< Res, L | forall{x: x in Res | not IsSameGenus(x, L) }, true >;

function FindLattice(i)
  x,phi,Aut:= ConstructLattice(Names[i] : Ext:= Ext[i]);
  x:= DiagonalJoin(-Matrix(Q, GramMatrix(x)), U);

  E:= [ e : e in ElementaryDivisors(Matrix(Z, x)) | e ne 1 ];
  d, P, n:= QuadraticFormInvariants(x);
  assert Ncols(x) le 19;
  assert n le 19;
  I:= {};
  for p in Set(PrimeDivisors( Z ! (2*d) )) join P  do
    s:= p in P select -1 else 1;
    if p eq 2 then s := -s; end if;
    if s eq -1 then Include(~I, p); end if;
  end for;

  F:= QuadraticFormWithInvariants(22-Ncols(x), -d, I, 19-n);
  test:= DiagonalJoin( x, F );
  dd, H, nn:= QuadraticFormInvariants(test);
  assert dd eq -1 and H eq {2} and nn eq 19;

  M:= MaximalIntegralLattice(Matrix(R, 1/2 * F));
  M:= Rescale(M, 2);
  d:= Z ! Determinant(x);
  ok, dM:= IsPrincipal(Volume(M));
  dM:= Abs(Z ! Q ! dM);
  PP:= Set(PrimeDivisors( d div dM ));
  for s in PP diff {2} do
    p:= BaseRing(M) * s;
    Es:= [ Valuation(e, s): e in E | Valuation(e,s) ne 0 ];
    S:= Sublattices(M, p, Valuation(d div dM, s) div 2 : CallBack:= Filter); 
    S:= [ L: L in S | Es eq ValDiv(L, s) ];
    if #S ne 1 then
      N:= Norms( LatticeModule( Matrix(R, -x) ), s);
      NN:= [ Norms(L, s): L in S ];
      I:= Indices(NN, N);
      assert #I eq 1;
      M:= S[I[1]];
    else
      M:= S[1];
    end if;
  end for;
  if 2 in PP then
    E2:= [ Valuation(e, 2): e in E | Valuation(e, 2) ne 0 ];
    p:= 2*BaseRing(M);
    v:= Valuation(d div dM, 2);
    v div:= 2;
    S:= [ L: L in MaximalSublattices(M, p : CallBack:= Filter) ];
    if v ne 1 then
      assert v eq 2;
      T:= [];
      for L in S, LL in MaximalSublattices(L, p : CallBack:= Filter) do
	if forall{x:x in T | not IsLocallyIsometric(LL, x, p)} then Append(~T, LL); end if;
      end for;
      S:= T;
    end if;
    S:= [ L: L in S | E2 eq ValDiv(L, 2) ]; assert #S ne 0;
    N:= Norms( LatticeModule( Matrix(R, -x) ), 2);
    NN:= [ Norms(L, 2): L in S ];
    I:= Indices(NN, N);
    assert #I eq 1;
    M:= S[I[1]];
  end if;
  assert Norms( LatticeModule( Matrix(R, -x) ), 0) eq Norms(M, 0);
  M:= LLLGram(Matrix(Q, GramMatrix(M)));
  return M, phi, Aut;
end function;

function Reflection(v, f)
  B:= Basis(Generic(Parent(v)));
  return Matrix([ b - 2* (v*f, b)/(v*f,v) * v : b in B ]);
end function;

function DecomposeAut(B, f)
  B:= Matrix(Q, B);
  f:= Matrix(Q, f);

  // Wir folgen Kneser, "Quadratische Formen" S.12f
  // Das geht zwar anders besser, ist aber komplizierter.

  _, T:= OrthogonalizeGram(f);
  n:= Nrows(f);
  Y:= Parent(B) ! 1;
  X:= [];
  for i in [1..n] do
    g:= T[i] * B * Y^-1;
    x:= T[i]-g;
    if (x*f,x) ne 0 then
      Y:= Reflection(x, f) * Y;
      Append(~X, x);
    else
      Y:= Reflection(T[i]+g, f) * Reflection(g, f) * Y;
      Append(~X, g);
      Append(~X, T[i]+g);
    end if;
  end for;
  Reverse(~X);

  assert (#X eq 0 and IsOne(B)) or (#X ne 0 and B eq &*[ Reflection(v, f) : v in X ]);
  return X;
end function;

// Zassenhaus-Trick:
function Zassenhaus(a, f)
  a:= Matrix(Q, a);
  b:= a+1;
  I:= Image(b);
  while I ne Image(b^2) do b:= b^2; I:= Image(b); end while;
  m:= Dimension(I);
  BI:= BasisMatrix(I);
  BK:= KernelMatrix(b);
  n:= Determinant(BK * f * Transpose(BK));
  B:= VerticalJoin( BI, BK );
  n *:= Determinant( Submatrix(BI * (a+1)/2 * B^-1, 1,1, m, m) ); 
  n:= Squarefree( Denominator(n) * Numerator(n) );
  return n;
end function;

function SpinorNorm(a,f)
  s:= Sign(&*[ Q | -(x*f,x) : x in DecomposeAut(a,f) ]);
  assert s eq Sign(Zassenhaus(a, -f));
  return s;
end function;

function FindIso(F, f, phi)
  D1:= Codomain(f);
  D2:= Codomain(phi);
  assert Ngens(D1) eq Ngens(D2);

  n1:= Mod2Z( (x*F,x) ) where x:= D1.1 @@ f;
  n2:= Mod2Z( (x*F,x) ) where x:= D1.2 @@ f;
  s12:= ModZ( ((D1.1 @@ f)*F, D1.2 @@ f) );

  I1:= [ x : i in D2 | Mod2Z( (x,x) ) eq n1 where x:= i @@ phi ];
  I2:= [ x : i in D2 | Mod2Z( (x,x) ) eq n2 where x:= i @@ phi ];
  J:= [ <x,y> : x in I1, y in I2 | ModZ( ( x, y ) ) eq s12 ];
  assert #J ne 0;

  if Ngens(D2) eq 3 then
    n3:= Mod2Z( (x*F,x) ) where x:= D1.3 @@ f;
    s13:= ModZ( ((D1.1 @@ f)*F, D1.3 @@ f) );
    s23:= ModZ( ((D1.2 @@ f)*F, D1.3 @@ f) );
    I3:= [ x : i in D2 | Mod2Z( (x,x) ) eq n3 where x:= i @@ phi ];
    ok:= exists(J){ < x[1], x[2], y> : x in J, y in I3 | ModZ( (y, x[1]) ) eq s13 and ModZ( (y, x[2] ) ) eq s23 };
    assert ok;
  else
    J:= J[1];
  end if;

  g:= iso < D1 -> D2 | [ x @ phi : x in J ] >;
  return g;
end function;

function MyGenus(X, h)
  X:= MyStabGens(X, func< d, g | Determinant(g @ h) * d >, 1);
  X:= sub< Universe(X) | X >;
  return #Indices(AbelianInvariants( AbelianQuotient(X) ), 0)/2;
end function;

function GuessSignature(G)
  G:= Simplify(G);
  E:= AbelianInvariants( AbelianQuotient(G) );
  g:= #Indices(E, 0);
  assert IsEven(g);
  g:= g div 2;
  E:= [ e: e in E | e ne 0 ];
  Cands:= [ [2,2,2,4], [2,2,2,2,2], [2,4,6], [2,2,2,3] ];
  for EE in Cands do
    n:= #EE;
    M:= VerticalJoin(DiagonalMatrix(EE), Matrix(n, [1^^n]));
    if [ e: e in ElementaryDivisors(M) | e ne 1] ne E then continue; end if;
    F:= FreeGroup(n + 2*g);
    R:= [ F.i^EE[i]: i in [1..n] ];
    Append(~R, &*[F | F.i : i in [1..n]] * &*[ F | (F.i, F.(i+1)) : i in [n+1..n+2*g by 2] ]);
    H:= quo< F | R>;
    if E ne ElementaryDivisors( AbelianQuotient(G) ) then continue; end if;
    if SearchForIsomorphism(G, H, 8) then
      return g, EE;
    end if;
  end for;
  error "add more cases";
end function;

Latex1:= "";
Latex2:= "";
for i in [1..25] do
  "========================", i;
  Names[i];
  label:= i+23;
  if label eq 38 then label:= "38/39";
  elif label gt 38 then label +:= 1;
  end if;
  G, phi, Diagramm := FindLattice(i);

  // Get generators for the automorphism group.
  P:=GetPerfectPoints(Matrix(Z, G) : Raw);
  // The stabilizers always lack -I_3, so we add it here.
  P`Stabilizers := [ sub< Generic(S) | S, -S.0> : S in P`Stabilizers ];
  HH, hh:= Presentation(P : simplify);
  ok:= exists(MinusOne){ x[2] : x in P`StabilizerPreimages | IsOne( -Matrix(x[1]) ) }; assert ok;

  I, f, DA:= DiscAut(G);
  "#O(q_T):", #DA;

  // Triviale Operation auf L^#/L:
  B:= [ I.i: i in [1..Ngens(I)] ];
  OTilde, Orbit:= MyStabGens( HH, func< o, g | [ ((b@@f)*Matrix(Q, g@hh))@f: b in o] >, B );
  OTilde:= sub< HH | ReduceGenerators(sub< HH | OTilde>) >;
  "[O(T) : O~(T)]=", #Orbit;
  assert #Orbit eq Index(HH, OTilde);

  spinor:= func< n, a | Sign(SpinorNorm(a @ hh, G)) * n >;
  OPlus, PM1:= MyStabGens(HH, spinor, 1);
  OPlus:= sub< HH | ReduceGenerators(sub< HH | OPlus >) >;
  assert #PM1 eq 2;
  temp, OrbitPlus:= MyStabGens( OPlus, func< o, g | [ ((b@@f)*Matrix(Q, g@hh))@f: b in o] >, B );

  idx:= exists{g: g in Generators(OTilde) | SpinorNorm(g @ hh, G) eq -1} select 2 else 1;
  assert #Orbit * idx eq 2*#OrbitPlus;
  assert 2*#OrbitPlus eq Index(HH, sub< HH | temp >);
  "[O~(T):O~+(T)]=", idx;
  "[O^+(T) : O~+(T)]=", #OrbitPlus;

  OTildePlus:= MyStabGens([ HH ! g : g in Generators(OTilde) ], spinor, 1);
  //assert sub< HH | OTildePlus> eq OTilde meet OPlus;
  //assert sub< HH | OTildePlus> eq sub< HH | temp >;
  temp:= Rewrite(HH, OPlus);		// get presentation for O^+
  assert MyGenus(OPlus, hh) eq 0;

  // Now we already know a presentation of OPlus. It therefore helps tremendeously
  // if we write OTildePlus as a subgroup of OPlus (of index at most 2)!
  OTildePlus:= sub< OPlus | ReduceGenerators(sub< OPlus | OTildePlus >) >;
  genus:= MyGenus( sub< OPlus | OTildePlus, MinusOne>, hh);
  "Genus: O~+(T):", genus;

  // Generating the 1st table in the appendix:
  Latex1 *:= Sprintf("%o & %o & %o & %o & %o\\\\ \n", label, LatexName(Names[i]), #OrbitPlus, idx, genus);

  if #Diagramm ne #DA then
    // This is ridiculuous. We need to enumerate (the basis images of) the group elements:
    X:= Codomain(phi);
    B:= [ X.i : i in [1..Ngens(X)]];
    _, DiagrammSet:= MyStabGens( Diagramm, func< o, a | [ a(x): x in o ] >, B );
    psi:= FindIso(G, f, phi);
    g:= f * psi;

    // Now we have an epimorphism g: T -> X:= M^/M.
    assert #DA eq #Orbit;

    GammaPlus, GammaPlusAut, GammaPlusPaths:= MyStabGens( OPlus, func< o, a |  [ ((x @@ g) * Matrix(Q, a @ hh)) @ g : x in o ]>, B);
    GammaPlus:= GammaPlus cat [ GammaPlusPaths[i] : i in [1..#GammaPlusPaths] | GammaPlusAut[i] in DiagrammSet ];
    GammaPlus:= sub< OPlus | ReduceGenerators(sub< OPlus | GammaPlus >) >;
    Gamma, GammaAut, GammaPaths:= MyStabGens( HH, func< o, a |  [ ((x @@ g) * Matrix(Q, a @ hh)) @ g : x in o ]>, B);
    Gamma:= Gamma cat [ GammaPaths[i] : i in [1..#GammaPaths] | GammaAut[i] in DiagrammSet ];
    Gamma:= sub< HH | ReduceGenerators(sub< HH | Gamma >) >;
    "[O^+(T):Gamma^+]= ", Index( OPlus, GammaPlus );
    "[Gamma^+:O~+(T)]= ", Index( GammaPlus, OTildePlus );
    "[Gamma:Gamma^+]= ", Index( Gamma, GammaPlus );
    genus:= MyGenus( sub< OPlus | GammaPlus, MinusOne>, hh);
    "Genus: Gamma^+:", genus;

    // Generating the 2nd table in the appendix:
    Latex2 *:= Sprintf("%o & %o & %o & %o & %o\\\\ \n", label, LatexName(Names[i]), Index( OPlus, GammaPlus ), Index(Gamma, GammaPlus), genus);
  end if;

/*
  X:= MyStabGens( OPlus, func< d, g | Determinant(g @ hh) * d >, 1);
  Invariants(AbelianQuotient(sub< OPlus | X>));
  _:= GuessSignature(sub< OPlus | X >);
 */
end for;

